import typing
from typing import Any, Dict, Optional, Union, cast

from rich.console import Console, ConsoleOptions, RenderResult
from rich.tree import Tree

from .protocol import TrioNursery, TrioTask
from .registry import RegisteredSCInfo, SCRegistry

""" Description Tree

    We trace the current system state by keeping our own task tree

    This tree must support:
    1. Build up a state tree based on a TrioTask
    2. Given a task, find it's parent
"""


TrioNode = Union[TrioNursery, TrioTask]


class DescNode:
    def __init__(self, node: TrioNode, registry: SCRegistry):
        self.parent: Optional[DescNode] = None
        self.children: typing.List[DescNode] = []
        self._registry: SCRegistry = registry

        # ref to the actual node
        self.ref: typing.Union[TrioTask, TrioNursery] = node

    @property
    def info(self) -> RegisteredSCInfo:
        return self._registry.get_info(self.ref)

    def __repr__(self):
        return f"<DescNode:{self.info.name}: {self.children}>"

    def _rich_node(self, parent: Tree):
        cur_node = parent.add(f"[yellow] name: {self.info.name}")
        [n._rich_node(parent=cur_node) for n in self.children]


class DescTree:
    def __init__(self, root, registry: Optional[SCRegistry]):
        self.root: DescNode = root

        reg: Any = SCRegistry() if registry == None else registry
        self._registry: SCRegistry = reg

        self.ref_2node: Dict[Union[TrioTask, TrioNursery], DescNode] = {}

        # only used for debugging
        self._nodes: Dict[str, DescNode] = {}

    @property
    def registry(self):
        return self._registry

    def node_by_name(self, name: str) -> Optional[DescNode]:
        """Get a description node by its name
        only used for debugging
        """
        return self._nodes.get(name, None)

    def __repr__(self):
        return f"{self.root}"

    @classmethod
    def build(
        cls, root_task: TrioTask, registry: Optional[SCRegistry] = None
    ) -> "DescTree":
        """Build a tree of from source"""

        _registry: SCRegistry = SCRegistry() if registry is None else registry

        root_desc = DescNode(root_task, _registry)

        tree = cls(root=root_desc, registry=_registry)

        def register(node: DescNode):
            name = tree._registry.get_name(node.ref)
            tree.ref_2node[node.ref] = node

            # only used for debugging
            tree._nodes[name] = node

        def build_task(task_desc: DescNode):
            task = cast(TrioTask, task_desc.ref)
            # build for it's child nurseries
            for n in task.child_nurseries:
                child_desc = DescNode(node=n, registry=_registry)
                child_desc.parent = task_desc
                task_desc.children.append(child_desc)
                register(child_desc)
                build_nursery(child_desc)

        def build_nursery(nursery_desc: DescNode):
            nursery = cast(TrioNursery, nursery_desc.ref)
            # build for it's child tasks
            for t in nursery.child_tasks:
                child_desc = DescNode(node=t, registry=_registry)
                child_desc.parent = nursery_desc
                nursery_desc.children.append(child_desc)
                register(child_desc)
                build_task(child_desc)

        register(root_desc)
        build_task(root_desc)
        return tree

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        """Support Python `rich` lib"""
        root = Tree("DescTree")
        self.root._rich_node(parent=root)
        yield root

    def get_parent_ref(self, target: TrioNode) -> Optional[TrioNode]:
        """Directly retrieve the parent nursery of a task"""
        node = self.ref_2node.get(target, None)
        if node is not None and node.parent is not None:
            return node.parent.ref
        return None

    def ensure_node_in_tree(self, node: DescNode):
        if node != self._nodes[node.info.name]:
            raise RuntimeError(f"bug: node not exists in tree: {node}")
        if self.ref_2node.get(node.ref, None) is None:
            raise RuntimeError(f"bug: ref not exists in tree: {node.ref}")

    def ensure_node_not_in_tree(self, node: DescNode):
        if node in self._nodes:
            raise RuntimeError(f"bug: node already exists in tree: {node}")
        if self.ref_2node.get(node.ref, None) is not None:
            raise RuntimeError(f"bug: ref already exists in tree: {node.ref}")

    def remove_ref(self, target: TrioNode):
        node = self.ref_2node.get(target, None)
        if node == None:
            raise RuntimeError("Bug: remove unexisting ref from registry")
        node = cast(DescNode, node)
        if len(node.children) > 0:
            raise RuntimeError("Child remains, cannot remove")

        # remove node from its parent
        if node.parent:
            node.parent.children.remove(node)
            node.parent = None
        # remove node from registry
        self._registry.remove(node.ref)

    def remove_node(self, node: DescNode):
        """Remove node and all of it's children from Desc tree,
        also break the link to it's parent"""

        print(f"[remove] receive node: {node}")
        self.ensure_node_in_tree(node)
        self.ref_2node.pop(node.ref)
        self._nodes.pop(node.info.name)

        # unregister all children in the tree
        for child in node.children:
            self.remove_node(child)

        if node.parent:
            node.parent.children.remove(node)
            # if we remove the only task from parent nursery,
            # the parent nursery should also be removed automatically
            # sinc Trio only notify us the creation/removement of tasks, but not nurseries
            if len(node.parent.children) == 0:
                self.remove_node(node.parent)
            node.parent = None

        self._registry.remove(node.ref)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['trio_vis']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.2.0,<22.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'rich>=9.13.0,<10.0.0',
 'trio-typing>=0.5.0,<0.6.0',
 'trio>=0.19.0,<0.20.0',
 'typing-extensions>=3,<4']

setup_kwargs = {
    'name': 'trio-vis',
    'version': '0.1.0',
    'description': 'Structured Concurrency visualizer for trio',
    'long_description': "# trio-vis\n\n`trio-vis` is a plugin for visualizing the scope history of your Trio project.\n\n![showcase](res/showcase.png)\n\n## How to use\n\n[sc-vis]: https://ianchen-tw.github.io/sc-vis\n[trio-vis-pip]: https://pypi.org/project/trio-vis/\n\n1. Install `trio-vis` via `pip install trio-vis` ([trio-vis-pip])\n2. In your source code, register `SC_Monitor()` as an Instrument while running `trio`\n\n    ```python\n    from trio_vis import SC_Monitor\n    trio.run(my_main_funciton, instruments=[SC_Monitor()])\n    ```\n\n3. After your program finished(or exited), the scope history would be stored in `./sc-logs.json`\n4. Upload your log file to [sc-visualizer][sc-vis], this is a twin project which focuses on visualization work.\n5. See your visualization result and help us improve.\n\n## Configuration\n\nImport `VisConfig` from `trio_vis`, and provide it as an argument while making your `SC_Monitor` object.\n\n```python\nfrom trio_vis import SC_Monitor, VisConfig\ncfg = VisConfig(print_task_tree=True)\ntrio.run(my_main_funciton, instruments=[SC_Monitor(config=cfg)])\n```\n\n## What does it do\n\n[ins-api]: https://trio.readthedocs.io/en/stable/reference-lowlevel.html#instrument-api\n\n`trio-vis` utilize the [Instrument API][ins-api] to monitor the lifetime of scopes (`Task`,`Nursery`).\nSince the [Instrument API][ins-api] doesn't provide callbacks for `Nursery`, we make inferences on our own.\n\n## Why visualize\n\n[trio]: https://github.com/python-trio/trio\n[trio-issue-413]: https://github.com/python-trio/trio/issues/413\n\n[curio]: https://github.com/dabeaz/curio\n[curio-monitor]: https://github.com/dabeaz/curio/blob/master/curio/monitor.py\n\nDerived from [curio], [trio] combines the idea of Structured Concurrency with existing single-threaded event-driven architecture. Which does make concurrent programs more manageable.\n\nTo make trio comparable with curio, contributors of trio also want to mimic the feature of [curio-monitor] to monitor the current system running state. This idea could be traced back to [trio-issue-413].\n\nSince then, projects have been developed (shown below).\n\nHowever, **trio is not curio**, at least lifetimes of scopes are structured by nature. I argue that by utilizing the feature of Structured Concurrency, we could visualize programs better.\nDevelopers could easily conceptualize their program, and bring their developing experience to the next level.\n\n### Previous work\n\n+ [python-trio/trio-monitor]: officail project developed under trio, however it use the old InstruementAPI\n+ [syncrypt/trio-inspector]: is a webmonitor to visualize the current state of the program\n+ [Tronic/trio-web-monitor]: a experiment to unified all previous work, developed by [Tronic](https://github.com/Tronic)\n+ [oremanj/trio-monitor]\n\n[python-trio/trio-monitor]:https://github.com/python-trio/trio-monitor\n[Tronic/trio-web-monitor]:https://github.com/Tronic/trio-web-monitor\n[syncrypt/trio-inspector]:https://github.com/syncrypt/trio-inspector\n[oremanj/trio-monitor]:https://github.com/oremanj/trio-monitor\n\n## Future plan\n\nThis project is in an early developing stage. Stay tuned for future update.\n",
    'author': 'Ian Chen',
    'author_email': 'ianre657@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ianchen-tw/trio-vis',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)



# Logger for structure-concurrent events
import atexit
import json
from abc import abstractmethod
from typing import Dict, List, Optional, cast

import attr
from typing_extensions import Protocol

from .desc_tree import TrioNode
from .registry import RegisteredSCInfo, SCRegistry


class Logger(Protocol):
    @abstractmethod
    def log_start(self, child: TrioNode, parent: Optional[TrioNode]):
        raise NotImplementedError

    @abstractmethod
    def log_exit(self, child: TrioNode, parent: Optional[TrioNode]):
        raise NotImplementedError


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SCEvent:
    time: int
    desc: str
    name: str
    type: str
    parent: Optional[str]

    def as_dict(self) -> Dict:
        # Parent need to be undefined if not specified
        return {k: v for k, v in attr.asdict(self).items() if v is not None}


def get_info(
    target: Optional[TrioNode], registry: SCRegistry
) -> Optional[RegisteredSCInfo]:
    if target == None:
        return None
    target = cast(TrioNode, target)
    return registry.get_info(target)


def scope_name(task_info: RegisteredSCInfo):
    if task_info.type != "task":
        raise RuntimeError("Can only call this method with task object")
    return f"{task_info.name}__TRIO_VIS_Tscope"


class SCLogger(Logger):
    def __init__(self, registry: SCRegistry, log_filename: str):
        self.event_id: int = 0
        self.registry: SCRegistry = registry
        self.events: List[Dict] = []
        self.log_filename: str = log_filename
        atexit.register(self._write_log)

    def _get_info(self, child, parent):
        if child is None:
            raise RuntimeError("target should not be none")
        child_info = get_info(child, self.registry)
        parent_info = get_info(parent, self.registry)
        return (child_info, parent_info)

    def log_start(self, child: TrioNode, parent: Optional[TrioNode]):
        child_info, parent_info = self._get_info(child, parent)
        if parent_info is None:
            # Root Scope
            self.events.append(
                SCEvent(
                    time=self.time,
                    name=scope_name(child_info),
                    desc="created",
                    parent=None,
                    type="scope",
                ).as_dict()
            )
            # Root task
            self.events.append(
                SCEvent(
                    time=self.time,
                    name=child_info.name,
                    desc="created",
                    parent=scope_name(child_info),
                    type="task",
                ).as_dict()
            )

            # print(f"Create root scope: {scope_name(child_info)}")
            # print(
            #     f"Create root task: {child_info.name} under scope:{scope_name(child_info)}"
            # )
        elif child_info.type == "task" and parent_info.type == "nursery":
            # Scope for child task
            self.events.append(
                SCEvent(
                    type="scope",
                    time=self.time,
                    name=scope_name(child_info),
                    desc="created",
                    parent=parent_info.name,
                ).as_dict()
            )
            # Child task
            self.events.append(
                SCEvent(
                    type="task",
                    time=self.time,
                    name=child_info.name,
                    desc="created",
                    parent=scope_name(child_info),
                ).as_dict()
            )
            # print(
            #     f"Create scope: {scope_name(child_info)} under scope:{parent_info.name}"
            # )
            # print(
            #     f"Create task: {child_info.name} under scope:{scope_name(child_info)}"
            # )
        elif child_info.type == "nursery" and parent_info.type == "task":
            self.events.append(
                SCEvent(
                    type="scope",
                    time=self.time,
                    name=child_info.name,
                    desc="created",
                    parent=scope_name(parent_info),
                ).as_dict()
            )
            # print(
            #     f"Create scope: {child_info.name} under scope: {scope_name(parent_info)}"
            # )
        else:
            print("spawn unknown")

    def log_exit(self, child: TrioNode, parent: Optional[TrioNode]):
        child_info, parent_info = self._get_info(child, parent)
        if parent_info is None:
            # root exit
            # print(f"Exit task: {child_info.name} under scope:{scope_name(child_info)}")
            # print(f"Exit scope: {scope_name(child_info)}")
            # Root task
            self.events.append(
                SCEvent(
                    time=self.time,
                    name=child_info.name,
                    desc="exited",
                    parent=scope_name(child_info),
                    type="task",
                ).as_dict()
            )
            # Root scope
            self.events.append(
                SCEvent(
                    time=self.time,
                    name=scope_name(child_info),
                    desc="exited",
                    parent=None,
                    type="task",
                ).as_dict()
            )
        elif child_info.type == "task" and parent_info.type == "nursery":
            # print(f"Exit task:{child_info.name} under scope:{scope_name(child_info)}")
            # print(f"Exit scope:{scope_name(child_info)} under scope:{parent_info.name}")
            self.events.append(
                SCEvent(
                    time=self.time,
                    name=child_info.name,
                    desc="exited",
                    parent=scope_name(child_info),
                    type="task",
                ).as_dict()
            )
            self.events.append(
                SCEvent(
                    time=self.time,
                    name=scope_name(child_info),
                    desc="exited",
                    parent=parent_info.name,
                    type="scope",
                ).as_dict()
            )

        elif child_info.type == "nursery" and parent_info.type == "task":
            # print(f"Exit scope:{child_info.name} under scope:{scope_name(parent_info)}")
            self.events.append(
                SCEvent(
                    time=self.time,
                    name=child_info.name,
                    desc="exited",
                    parent=scope_name(parent_info),
                    type="scope",
                ).as_dict()
            )
        else:
            print("exit unknown")

    def _write_log(self):
        sc_logs = {
            "config": {"makeDirectScopeTransparent": True},
            "runRecords": self.events,
        }
        with open(self.log_filename, "w") as log_file:
            json.dump(sc_logs, log_file, indent=4)

    @property
    def time(self) -> int:
        ret = self.event_id
        self.event_id += 1
        return ret


def log():
    event = SCEvent(2, "t0", "start", "n0")
    print(json.dumps(event.as_dict()))


def main():
    print("good")
    log()


if __name__ == "__main__":
    main()
